# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package MakeWorld::Package;

use strict;
use warnings;

use File::Path qw[ rmtree ];
use File::Spec::Functions qw[ catdir catfile path file_name_is_absolute ];

use Module::Pluggable require => 1, search_path => qw[ MakeWorld::Package ];

use MakeWorld::Check qw[ check %BOOLEAN ];

use Cwd qw/ realpath /;

use MakeWorld::Global;
use MakeWorld::Interpolate;
use MakeWorld::Log;
use MakeWorld::Run;
use MakeWorld::Stages;
use MakeWorld::Archive;
use MakeWorld::Utils qw[ find_exe ];

# environmental variables which are paths.  variables which end in PATH
# are automatically recognized
my %PathVar = map { $_ => 1 }
  qw[ TEXINPUTS PERLLIB PERL5LIB ];

my @PackageTypes =  grep { ! /::(Unknown|Custom)$/ } __PACKAGE__->plugins();
my %PackageTypes =  map { my $pkg = $_;
			  $pkg =~ s/^.*:://;
			  ( lc($pkg) => $_ ) } @PackageTypes;
$PackageTypes{custom} = 'MakeWorld::Package::Custom';

################################################################################
# create a generic Package object.  This object will later be reblessed into a
# Package subclass once the package type has been discovered.  Nothing can
# be done here which depends upon the package type.

sub new
{
    my ( $class, $name, %attr ) = @_;

    my $self = bless
	       { name   => $name,
		 archive_name => $attr{archive_name} // $name,
		 pkgspec => defined $attr{pkgspec} ? $attr{pkgspec} : undef,
		 tracked_files => [],
		 mogrify => [],
		 env => {},

		 # this will be redone once the package is specialized.
		 stage => MakeWorld::Stages->new( stages => MakeWorld::Stages::stagedef( $attr{stages}),
						  filter => $Opt->stages,
						),
	       }, $class;

    $self->set_option( $attr{option} ) if exists $attr{option};

    return $self;
}

################################################################################
# Generic initialization of a Package object *after* it has been specialized by
# being reblessed into the appropriate Package subclass.

sub init {
    my $self = shift;

    $self->{stage} =
      MakeWorld::Stages->new(
			     stages => $self->can('stages')
			                 ? $self->stages
			                 : MakeWorld::Stages::stagedef( $self->option('stages') ),
			     filter => $Opt->stages
			    );

    $self->{hook} = { map { $_ => { pre => [], post => [] } }
		      @{$self->{stage}{stages} } } ;

}


################################################################################
# determine if this package class supports the package type based upon the
# name of the archive file

sub matches_archive { my $archive = shift; return; }

################################################################################
# extract the associated archive.  this should be overridden by package classes
# if they expect to return true from the matches_archive() method

sub extract {

    my ( $self ) = @_;

    return $self->archive->extract( $self->logfile( 'extract' ), $self->ok_to_run( 'extract' ) );

}

################################################################################
# determine if this package class supports the package type based upon
# the contents of the current directory

sub matches_contents { die( "internal error" ); }


################################################################################
# validation template for stages.  package classes may override this

sub stage_vtpl {

    my ( $self, $stage ) = @_;

    return { exe => { required => 0 },
	     arg => { required => 0 },
	     placeholder => { default => 0 },
	   }
}




################################################################################

sub name            { $_[0]->{name}    }
sub archive_name    { $_[0]->{archive_name}    }

sub option  { my $self = shift;
	      @_ ?
		( exists $self->{option}{$_[0]} ? $self->{option}{$_[0]} : undef )
		  : $self->{option}
	      }
sub archive { $_[0]->{archive} }
sub base    { $_[0]->{archive}->base }
sub hook    { $_[0]->{hook} }

sub set_option {
    my ( $self, $options ) = @_;

    $self->{option} ||= { %{ $Opt->defoption } };

    for my $hash ( $options, $Opt->option )
    {
	$self->{option}{$_} = $hash->{$_} for keys %$hash;
    }

};

sub set_env{

    my $self = shift;

    my %env;
    while( 'HASH' eq ref $_[0] )
    {
	%env = ( %env, %{shift()} );
    }

    %env = ( %env, @_ );

    $self->{env}{$_} = $env{$_} for keys %env;
    return;

}

################################################################################
# Run the package specification recipe on a generic Package object.  Nothing
# which is specific to a type of Package (i.e. anything with stages) can be
# dealt with here.  It is put off until the process method is called.

sub apply_spec {

    my $self = shift;

    if ( defined $self->{pkgspec} )
    {
	my ( $name, $options ) = $self->{pkgspec}->parse( $self );

	die( "internal error: expected <package>, got <$name>\n" )
	  unless $name eq 'package';

	# delay some checks until after we know what type of package it is
	local $Params::Check::ALLOW_UNKNOWN = 1;

	my %tpl = (
		   name      => { required => 1 },
		   archive_name => { required => 0 },
		   pkg_type  => { required => 0,
				  allow => [ keys %PackageTypes ]
				},
		   matchre    => { required => 0,
				   allow => [ sub { eval "qr|$_[0]|; 1;" } ],
				 },
		   versionre => { required => 0,
				  allow => [ sub { eval "qr|$_[0]|; 1;" } ],
				},
		   version_extract => { required => 0,
				  allow => [ sub { eval "qr|$_[0]|; 1;" } ],
				},
		   cache    => { default => defined $Opt->cache,
				 %BOOLEAN,
			       },

		   exclude_if_inpath => { required => 0 },

		   exclude => { required => 0, %BOOLEAN },

		   libtool => { required => 0 },

		   stages => { required => 0,
			       allow => [ \&MakeWorld::Stages::legal_stage ],
			     },

		  );

	my $opts = check( \%tpl, $options );

	# remove the options that were validated and save the rest for later
	delete @{$options}{keys %tpl};
	$self->{options} = $options;

	$self->set_option( $opts );
    }

    return;
}


################################################################################
# validate a package recipe *after* it has been reblessed into the specific
# Package class

sub validate {

    my $self = shift;

    # now that the package type is known, finish validation of recipe
    # these are the options left over from initial validation of the package.
    my $opts = $self->{options};

    # temporary storage for newly validated options
    my %opts;

    # Params::Check is not reentrant, so can't do recursive checks
    for my $stage (  @{ $self->{stage}{stages} } )
    {
	if ( defined $opts->{$stage} ) {

	    die( "<$opts->{name}><$stage>: no attributes specified?\n" )
	      unless 'HASH' eq ref $opts->{$stage};

	    $opts{$stage} = check( $self->stage_vtpl( $stage ),
				   delete $opts->{$stage} );
	}

	else {

	    $opts{$stage} = check( $self->stage_vtpl( $stage ), {} );
	}
    }

    $self->set_option( \%opts );

    $self->{stage}->allow_stages( $self->option('stages') );

    $self->validate_mogrify_hooks( $_ ) foreach @{ $self->{mogrify} };

    if ( $self->can( 'vtpl' ) ) {
	my $opts = check( $self->vtpl, $self->{options} );
	$self->set_option( $opts );
    }

    return;
}

################################################################################
# set up environment for package

sub setenv {

    my $self = shift;

    # set package specific environment variables. variables with an empty
    # value should be deleted.  this is done twice; once here, and also
    # after merging with $Opt->define and $Opt->odefine.  FIXME: WHY?
    while( my ( $var, $value ) = each %{$self->{env}} )
    {
	if ( $value ne '' )
	{
	    $ENV{$var} = interpolate( $value, $self->archive );
	}
	else
	{
	    delete $ENV{$var};
	}
    }

    # merged environmental variables
    my $env = $self->option('env');
    while( my ( $var, $value ) = each %{$Opt->define} )
    {
	$value = interpolate( $value, $self->archive );

	if ( $env && exists $env->{$var} )
	{
	    $value = join( ($var =~ /PATH$/ || $PathVar{$var}) ? ':' : ' ',
			   $value, $env->{$var} );
	}
	$ENV{$var} = $value;
    }

    # general overrides
    while( my ( $var, $value ) = each %{$Opt->odefine} )
    {
	$ENV{$var} = interpolate( $value, $self->archive );
    }


    # package specific environmental variable deletions
    delete @ENV{
	map { $_ } grep { $self->{env}{$_} eq '' } keys %{$self->{env}}
    };

    return;
}



################################################################################
# find the archive

sub find_archive
{
    my $pkg = shift;
    $pkg->{archive} =
      MakeWorld::Archive->new( { name => $pkg->name,
				 archive_name => $pkg->archive_name,
				 map { $_ => $pkg->option($_) }
				 qw[ matchre versionre version_extract ],
		    },  @_ );
}

################################################################################
# process a package.  This:
#
# 1. extracts the package
# 2. determines its type
# 3. finishes validating the recipe
# 4. sets up environment variables
# 5. runs the various stages

sub process
{
    my $pkg = shift;

    my $name    = $pkg->name;

    # if true, check that package contents match package type
    # this is true only for explicitly defined package types
    my $check_contents = 0;

    # switch to the builddir
    chdir( $Opt->builddir )
      or die( 'unable to cd to ', $Opt->builddir, "\n" );

    # if one of the packages has claimed the archive,
    # extract it and rebless the package into its class
    if ( $pkg->archive->class ) {
	bless $pkg, $pkg->archive->class;
	msg( MakeWorld::Log::Debug, "Assigning to " . $pkg->archive->class );
    }

    # extract the archive into the builddir
    my $srcdir = $pkg->extract;

    # switch to the source dir if possible
    if ( -d $srcdir )
    {
	chdir( $srcdir )
	  or die( "unable to cd to $srcdir\n" );

	# if the package class hasn't yet been determined,
	# do it here.
	if ( ref $pkg eq __PACKAGE__ )
	{
	    # explicit specification of package type?  normally
	    # this is only done for custom types, but we handle 'em all
	    # anyway, 'cause we're that sort of folks.
	    if ( my $pkg_type = $pkg->option('pkg_type') )
	    {
		die( "unknown package type: $pkg_type\n" )
		  unless exists $PackageTypes{lc $pkg_type};
		my $pkgtype =  $PackageTypes{lc $pkg_type};

		# indicate that we need to verify that the contents are
		# correct for this package type (if not Custom)
		$check_contents = $pkgtype !~ /Custom/ ? 1 : 0;

		bless $pkg, $pkgtype;
	    }

	    # determine what type of package we're dealing with.
	    else
	    {
		my ( $pkgclass ) = grep { $_->matches_contents } @PackageTypes;

		$pkgclass //= 'MakeWorld::Package::Custom';

		# if we found something, rebless into that package, else
		# make it a Custom class and let it complain

		bless $pkg, $pkgclass;
		msg( MakeWorld::Log::Debug, "Assigning to $pkgclass" );
	    }
	}

    }

    # no dir and not a dryrun? this shouldn't happen
    elsif ( ! $Opt->dryrun )
    {
	die( "extraction to $srcdir failed?\n" );
    }

    # no srcdir && dryrun; don't know which package it is, so use a
    # do-nothing noisy class
    else
    {
	MakeWorld::Package::Unknown->new( $pkg );
    }

    $pkg->init;

    # now that the package type is known, finish validation of recipe
    $pkg->validate;

    # customize environment
    local %ENV = %ENV;
    $pkg->setenv;

    # personalize options by interpolating against the archive and
    # command line option values
    {
	my %popt;
	while( my ( $key, $value ) = each %{$pkg->option} ) {
	    next if $key eq 'env';
	    $popt{$key} = interpolate( $value, $pkg->archive );
	}
	$pkg->set_option( \%popt );
    }

    # explicitly do this as extract is a special stage and isn't
    # in the package stage list
    $pkg->prepost_stage( 'extract', 'post' );

    # check contents against package type if it is required
    die( "package does not meet requirements for package type ",
	 $pkg->option('pkg_type'), "\n" )
      if $check_contents && ! $pkg->matches_contents;


    # iterate over stages, performing requested actions
    for my $stage ( @{$pkg->{stage}{stages}} )
    {
	my $method;

	unless ( $method = $pkg->can( $stage ) ) {
	    msg( MakeWorld::Log::Notice,
		 "$stage is not supported for this package class; skipping" );
	    next;
	}

	$pkg->prepost_stage( $stage, 'pre' );

	if ( $pkg->ok_to_run( $stage, 'main' ) )
	{
	    msg( MakeWorld::Log::Notice, $stage );
	    my $pfx= MakeWorld::Log->new( $stage );
	    eval {
		my $rd = $pkg->logredir( $stage );
		$method->($pkg);
	    };
	    if ( my $err = $@ ) {
		my $logfile = $pkg->logfile($stage);
		die ( $err, defined $logfile ? "for more info see\n$logfile\n" : () );
	    }
	}

	$pkg->prepost_stage( $stage, 'post' );
    }

    # switch back to the builddir
    chdir( $Opt->builddir )
      or die( 'unable to cd to ', $Opt->builddir, "\n" );

    # delete the builddir
    rmtree( $srcdir )
      if $Opt->clean && ! $Opt->dryrun &&
              $pkg->{stage}->ok_to_run( $pkg->{stage}->last_stage );

    # remove any other files that are tracked
    if ( $Opt->clean && ! $Opt->dryrun )
    {
	unlink( $_ ) foreach @{$pkg->{tracked_files}};
    }
}

################################################################################
# run something before or after a stage

sub prepost_stage {

    my ( $pkg, $stage, $location ) = @_;

    my @path = ( $stage, $location );

    return unless $pkg->ok_to_run( @path );

    for my $mogrify ( @{ $pkg->hook->{$stage}{$location} } )
    {
	my $stage = join(':', @path, $mogrify->{name} );

	msg( MakeWorld::Log::Notice, $stage );
	my $pfx= MakeWorld::Log->new( $stage );

	# these options haven't yet been interpolated into
	my $opt = interpolate( $mogrify->{attr}, $pkg->archive );
	eval {
	    my $rd = $pkg->logredir( $stage );
	    $mogrify->{class}->mogrify($pkg, $opt );
	};
	if ( my $err = $@ ) {
	    my $logfile = $pkg->logfile($stage);
	    die( $err , defined $logfile ? "for more info see\n$logfile\n" : () );
	}
    }

    return;
}

################################################################################

sub make { main::run( $Opt->make_cmd, @_ ); }


################################################################################

sub ok_to_run {

    my $self = shift;

    return $self->{stage}->ok_to_run( @_ );

}

################################################################################

sub run_stage {

    my ( $self, $stage ) = @_;

    my $option = $self->{option}{$stage};

    return 1 if $option->{placeholder};

    return 0 unless $option->{exe};

    run( find_exe($option->{exe}), $option->{arg} )
	or die( "error in $stage stage\n" );

    return 1;
}


################################################################################
# it's up to the Package subclasses to perform actions based upon package type
# these are called by those subclass methods to do package specific
# actions.  they return 0 if they haven't handled the action.

sub configure
{
    return $_[0]->run_stage( 'configure' );
}

sub build
{
    return $_[0]->run_stage( 'build' );
}

sub test
{
    return $_[0]->run_stage( 'test' );
}

sub install
{
    return $_[0]->run_stage( 'install' );
}

#----------------------------------------------------------------#

sub is_excluded
{
    my ( $self ) = @_;

    return 1 if $self->{option}{exclude};

    if ( exists $self->{option}{exclude_if_inpath}{exe} )
    {
	return 1 if $self->find_in_path( $self->{option}{exclude_if_inpath}{exe} );
    }

    return 0;
}

#----------------------------------------------------------------#

sub find_in_path
{
    my ( $self, $exe ) = @_;

    my $bindir = realpath( catdir( $Opt->exec_prefix, 'bin' ) );

    foreach my $path ( path() )
    {
	next if realpath($path) eq $bindir;

	return 1 if -x catfile( $path, $exe );
    }

    return;
}

#----------------------------------------------------------------#




#----------------------------------------------------------------#

sub logfile
{
    my ( $self, $ext ) = @_;

    my $logfile;

    if ( $Opt->logredir )
    {
	$logfile = catfile( $Opt->logdir, $_[0]->name . "_$ext.log" );
	push @{$self->{tracked_files}}, $logfile;
    }


    return $logfile;
}

sub logredir
{
    my ( $self, $ext ) = @_;

    return MakeWorld::IORedir->new( file => $self->logfile( $ext ) );
}


################################################################################
#
# mogrification interface

# a package may provide transformation routines which will act upon the
# package directory at the various stages.  This is a somewhat kludgy means
# of providing verification and running of those stages

################################################################################
# called by the recipe parser to add a mogrification command to the
# package's list.

sub add_mogrify {

    my ( $self, $mogrify ) = @_;

    push @{ $self->{mogrify} }, $mogrify;

    return;
}

################################################################################
# validate that the requested hook is legit and add the command to the
# hook list.  called after a package has been specialized

sub validate_mogrify_hooks {

    my $self = shift;
    my $mogrify = shift;

    die( "invalid hook attribute value for $mogrify->{name}: $mogrify->{hook}\n" )
	unless $self->{stage}->is_legal_hook( $mogrify->{hook} );

    my ( $stage, $location ) = $self->{stage}->parse_hook( $mogrify->{hook} );

    push @{ $self->{hook}{$stage}{$location} }, $mogrify;

}

1;
